
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, cPickle
import constants, pathnames, utilities, mrquery
from LN import LN, MLLN

class DistanceDirectionLevel(object):
    """Distance, Direction, Level: Select pairs of directories to compare.

    See "levels.txt".
    """
    def __init__(self, topdatadir, basename, distance=None, direction=None,
          level=None):
        """Chooses which parts of the image directory tree will be used in "html.py".

        distance     'close', 'far' or None.
        direction    'upwards', 'downwards' or None.
        level        An int >= 0 or None.
        If all three are None, all images are used.
        The arguments must all be None or none of them be None.
        See "levels.txt".
        """
        self.distance = distance
        self.direction = direction
        self.level = level
        self.check_ddl_args()

        self.mrq = mrquery.FetchOrCreate_MRQuery(topdatadir, basename, 'as_is')
        print 'There are', len(self.mrq.sortlist), 'items in mrq.sortlist'
        self.critlevels = self.find_critlevels()
        self.make_ddl_sortlist()
        self.mrq.Store()

    def check_ddl_args(self):
        """Error checks for the arguments of __init__"""
        if self.distance not in ('close', 'far', None):
            raise TypeError("\"distance\" must be 'close', 'far' or None")
        if self.direction not in ('upwards', 'downwards', None):
            raise TypeError("\"direction\" must be 'upwards', "
                    "downwards' or None")
        if self.level is not None and (not isinstance(self.level, int) or
                      self.level < 0):
            raise TypeError("\"level\" must be None or an int >= 0")
        a = self.distance is None
        b = self.direction is None
        c = self.level is None
        if (a or b or c) and (not a or not b or not c):
            raise TypeError('"distance", "direction" and "level" '
                'must all be None or none of them be None')

    def find_critlevels(self):
        """Find the critical level for each image name.

        This is done here to save calculations in "html.py".
        """
        if self.distance is None:
            return None
        critlevels = dict()
        if self.direction == 'downwards':
            for hashcode in self.mrq.hash2ImData:
                for N in self.mrq.hash2ImData[hashcode]:
                    # Left side of inequality (2) in "levels.txt".
                    critlevels[N] = self.level
        else:
            for hashcode in self.mrq.hash2ImData:
                for N in self.mrq.hash2ImData[hashcode]:
                    # Left side of inequalities (1) in "levels.txt".
                    critlevels[N] = N.level - self.level
        return critlevels

    def levelsOK(self, N1, N2):
        """Do two names satisfy the distance-direction-level criteria"""
        if self.distance is None:
            return True
        name1 = N1.virtual_name
        name2 = N2.virtual_name
        prefix = os.path.commonprefix( (name1, name2) )
        i = prefix.rindex('/')
        prefix = prefix[:i]
        len1 = name1.count('/') - prefix.count('/') - 1
        len2 = name2.count('/') - prefix.count('/') - 1

        # Inequalities (1) in "levels.txt".
        i = prefix.count('/') + 1
        b = (self.critlevels[N1] < i and self.critlevels[N2] < i)
        if self.distance == 'far':
            return not b
        return b

    def make_ddl_sortlist(self):
        """Select from mrq.sortlist pairs of names that satisfy the distance-direction-level criteria"""
        ddl_sortlist = list()
        # Start with sets of identical images:
        for hashcode in self.mrq.hash2ImData:
            ID = self.mrq.hash2ImData[hashcode]
            if len(ID) > 1:
               for i in range(len(ID) - 1):
                   for j in range(i+1, len(ID)):
                       ddl_sortlist.append( (1.0, ID[i], ID[j], ID, ID) )

        # Do the rest.
        for score, imnum1, imnum2 in self.mrq.sortlist:
            hashcode1 = self.mrq.imnum2hash[imnum1]
            hashcode2 = self.mrq.imnum2hash[imnum2]
            ID1 = self.mrq.hash2ImData[hashcode1]
            ID2 = self.mrq.hash2ImData[hashcode2]
            for N1 in ID1:
                for N2 in ID2:
                    if self.levelsOK(N1, N2):
                        ddl_sortlist.append( (score, N1, N2, ID1, ID2) )
        self.mrq.ddl_sortlist = ddl_sortlist
