
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import traceback, textwrap, inspect, sys

def LN(*args, **kwargs):
    """Prints a line number and some text.

    A variable number of positional arguments are allowed. If
        LN(obj0, obj1, obj2)
    is called, the text part of the output looks like the output from
        print obj0, obj1, obj2
    The optional keyword "wrap" causes the message to be line-wrapped. The
    argument to "wrap" should be True or False; the default is False. "name"
    is another optional keyword parameter which is best explained by an
    example:
        from linenum import LN, MLLN
        def fun1():
            print LN('error', 'is', 'here')
        def fun2():
            print LN('error',  'is', 'here', name='mess')
        fun1()
        fun2()
    The output is:
        L3 fun1: error is here
        L5 mess: error is here
    """
    sys.stdout.flush()
    sys.stderr.flush()
    stack = traceback.extract_stack()
    path, lineno, funcname, source_line = stack[-2]
    modulename = inspect.getmodulename(path)
    out = []
    for obj in args:
        out.append(str(obj))

    text = ' '.join(out)
    if 'name' in kwargs:
        name = kwargs['name']
    else:
        name = '%s, %s' % (modulename, funcname)
    # The output will be prettier if the file has < 1000 lines.
    text = 'L%-3s %s: %s' % (lineno, name, text)
    if 'wrap' in kwargs and kwargs['wrap']:
        text = textwrap.fill(text)
    print >> sys.stdout, text
    sys.stdout.flush()

def MLLN(*args, **kwargs):
    """Prints a line number and some text. Each argument is output on a separate line.

    A variable number of positional arguments are allowed.

    The optional keyword "wrap" causes the message to be line-wrapped. The
    argument to "wrap" should be "1" or "True". "name" is another optional
    keyword parameter which is best explained by an example (mess.py):
        from linenum import LN, MLLN
        def fun1():
            print MLLN('error', 'is', 'here')
        def fun2():
            print MLLN('error',  'is', 'here', name='messy')
        fun1()
        fun2()
    The output is:
        L3   mess, fun1:
          error

          is

          here

        L5   messy:
          error

          is

          here
    """
    sys.stdout.flush()
    sys.stderr.flush()
    stack = traceback.extract_stack()
    path, lineno, funcname, source_line = stack[-2]
    modulename = inspect.getmodulename(path)
    out = []
    for obj in args:
        text = str(obj)
        lines = text.splitlines()
        lines.append('')
        text = '  ' + '\n  '.join(lines)
        out.append(text)
    text = '\n'.join(out)
    if 'name' in kwargs:
        name = kwargs['name']
    else:
        name = '%s, %s' % (modulename, funcname)
    # The output will be prettier if the file has < 1000 lines.
    text = 'L%-3s %s:\n%s' % (lineno, name, text)
    if 'wrap' in kwargs and kwargs['wrap']:
        text = textwrap.fill(text)
    print >> sys.stdout, text
    sys.stdout.flush()
