
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import sys, os, zlib, cPickle
import Image, numarray
from numarray.numerictypes import *
import clipcast
"""
IMage processing system. "IM" is pronounced "I'm".

The original form of IM wrapped, in a symmetric manner, Numeric, PIL, and
the Intel OpenCV system
(http://www.intel.com/research/mrl/research/opencv/). Currently the system
is tilted toward numarray with PIL somewhat in the background. I removed
OpenCV because it is very nasty to wrap and I didn't use it much. I
downplayed PIL because I found Numeric / numarray pleasant to use at both
the Python and C levels.

NOTES:
For numarrays, I am putting the bands in the first dimension so that
  R, G, B = array[0], array[1], array[2]

Order of Coordinates
PIL uses (x, y) coordinates starting in the upper left corner while numarray
uses (r, c) coordinates starting in the same place. Therefore if the size of
a PIL image is (width, height) then the shape of the corresponding numarray
is (height, width). If the shape of a numarray is (rows, cols) then the size
of the corresponding PIL image is (cols, rows).

In function calls, height, width, bands are used in that order.

"Ntype" means any of the types imported from numerictypes.

"NStype" means any of the objects or strings used to tell numarray a type.
   For example, "numarray.array([1,2,3], 'UInt8')".

"mode" or "Ptype" mean any of PIL's modes.

"NPtype" can be any Ntype or any Ptype.

"NSPtype" can be any NStype or any Ptype.

Because "1" and "F" are both numarray codes and PIL modes, I am not allowing
the use of the Numeric codes "1", "s", "i", "u", "f", "d", "F", or "D".
"""
class CommonErrors(Exception):
    pass
class ComplexNotImplemented(CommonErrors):
    pass
class PilModeOneNotImplemented(CommonErrors):
    pass
class UnsafeCast(TypeError):
    pass

#> === NA_types ===================================================

class MODULE:
    NUMARRAY = 0
    PIL = 1

MODULEname_tuple = ('MODULE.NUMARRAY', 'MODULE.PIL')

def ModuleName(module):
    return MODULEname_tuple[module]

PIL_Modes = ('1', 'L', 'P', 'I', 'F', 'RGB', 'RGBA', 'CMYK', 'YCbCr')

NA_Types = (Bool, Int8, UInt8, Int16, UInt16, Int32, UInt32, Int64,
            UInt64, Float32, Float64, Complex32, Complex64)

NA_BadCodes = ('1', 's', 'i', 'u', 'f', 'd', 'F', 'D')

NA_TypeNameStrings = numarray.numerictypes.typeDict.keys()

def PIL_ModeCheck(mode):
    """Raises a TypeError if input not a PIL mode."""
    if not isinstance(mode, str) or mode not in PIL_Modes:
        raise TypeError, '"%s" is not a PIL mode' % mode

def NA_GetType(NStype):
    """Converts a numarray type specifier to a numarray type object."""
    return numarray.zeros(tuple(), NStype).type()

def NA_TypeCheck(NStype):
    """Is NStype one of the numarray type specifiers?"""
    if str(NStype) in NA_BadCodes:
        raise ValueError, 'IM forbids use of Numeric one character type codes'
    NA_GetType(NStype)

def PIL_IsMode(mode):
    """Is input a PIL mode?"""
    return mode in PIL_Modes

def TypeCheck(NSPtype):
    """Is the input a NumericType, a numarray string type name, or a PIL mode?"""
    if NSPtype in PIL_modes:
        return True
    try:
        NA_TypeCheck(NSPtype)
    except ValueError:
        raise TypeError, '"%s" is not a legal NSPtype' % NSPtype
    except TypeError, s:
        if str(s) == 'Not a numeric type':
            raise TypeError, '"%s" is not a legal NSPtype' % NSPtype

def NA_TypeName(NStype):
    """Returns a name for a NumericType object."""
    return str(NA_GetType(NStype))

def NA_CheckShape(arr):
    """Raises an exception if an entry in shape is <= 0"""
    for n in arr.shape:
        if n < 1:
            raise ValueError, 'entry in shape is <= 0'

def NA_GetTypeFromTypeno(typeno):
    """Converts a numarray type number into a numarray type"""
    if typeno < 1 or typeno > 13:
        raise IndexError('1 <= typeno <= 13 is required')
    return NA_Types[typeno - 1]

def NA_TypenoFromType(Ntype):
    """Converts a numarray type or typecode into a numarray type number"""
    return list(NA_Types).index(Ntype) + 1

#--- PIL ---------------------------------------------------------------

# I don't understand mode "1".

PIL_ModeToType = { '1' : Bool, 'L' : UInt8, 'P' : UInt8, 'I' : Int32,
    'F' : Float32, 'RGB': UInt8, 'RGBA': UInt8, 'CMYK': UInt8,
    'YCbCr': UInt8 }

PIL_BandsPerMode = { '1' : 1, 'L' : 1, 'P' : 1, 'I' : 1,
    'F' : 1, 'RGB': 3, 'RGBA': 4, 'CMYK': 4,'YCbCr': 3 }

NA_To_PIL_1band = { Bool : '1', UInt8 : 'L', Int8 : 'I', Int16 : 'I',
      UInt16 : 'I', Int32 : 'I', Float32 : 'F' }

def IsArray(arrin):
    """Is the array one of the types allowed?"""
    return isinstance(arrin, Image.Image) or \
           isinstance(arrin, numarray.NumArray)

#--- general -----------------------------------------------------

# These are useful for testing where symmetry between numarray and PIL can
# be useful.

def TypeName(module, NSPtype):
    if module == MODULE.NUMARRAY:
        return NA_TypeName(NSPtype)
    elif module == MODULE.PIL:
        PIL_ModeCheck(NSPtype)
        return NSPtype
    else:
        raise ValueError, 'Illegal module type'

def GetPixel(array, pixel):
    module = Module(array)
    if module == MODULE.PIL:
        return array.getpixel( (pixel[1], pixel[0]) )
    elif module == MODULE.NUMARRAY:
        return array[pixel]

#--- safe casts --------------------------------------------------

# Safe coercions according to numarray.
NA_UpwardsTo_NA = {
    Bool   : NA_Types,
    Int8   : (Int8, Int16, Int32, Int64, Float32, Float64, Complex32, Complex64),
    UInt8  : NA_Types[2:],
    Int16  : (Int16, Int32, Int64, Float32, Float64, Complex32, Complex64),
    UInt16 : NA_Types[4:],
    Int32  : (Int32, Int64, Float32, Float64, Complex32, Complex64),
    UInt32 : NA_Types[6:],
    Int64  : (Int64, Float32, Float64, Complex32, Complex64),
    UInt64 : NA_Types[8:],
    Float32 : (Float32, Float64, Complex32, Complex64),
    Float64 : (Float64, Complex64),
    Complex32 : (Complex32, Complex64),
    Complex64 : (Complex64,) }

NA_UpwardsTo_PIL = {
    Bool  : PIL_Modes,
    Int8  : ('I', 'F'),
    UInt8 : ('L', 'P', 'I', 'F', 'RGB', 'RGBA', 'CMYK', 'YCbCr'),
    Int16 : ('I', 'F'), UInt16 : ('I', 'F'),
    Int32 : ('I', 'F'), UInt32 : ('F',),
    Int64 : (),
    UInt64 : (),
    Float32 : ('F',),
    Float64 : (),
    Complex32 : (),
    Complex64 : () }

PIL_UpwardsTo_PIL = {
    '1'     : PIL_Modes,
    'L'     : ('L', 'P', 'I', 'F', 'RGB', 'RGBA', 'CMYK', 'YCbCr'),
    'P'     : ('L', 'P', 'I', 'F', 'RGB', 'RGBA', 'CMYK', 'YCbCr'),
    'I'     : ('I', 'F'),
    'F'     : ('F'),
    'RGB'   : ('RGB', 'RGBA', 'CMYK', 'YCbCr'),
    'RGBA'  : ('RGBA',),
    'CMYK'  : ('RGB', 'RGBA', 'CMYK', 'YCbCr'),
    'YCbCr' : ('RGB', 'RGBA', 'CMYK', 'YCbCr') }

PIL_UpwardsTo_NA = {
    '1'     : NA_Types,
    'L'     : NA_Types[2:],
    'P'     : NA_Types[2:],
    'I'     : (Int32, Int64, Float32, Float64, Complex32, Complex64),
    'F'     : (Float32, Float64, Complex32, Complex64),
    'RGB'   : NA_Types[2:],
    'RGBA'  : NA_Types[2:],
    'CMYK'  : NA_Types[2:],
    'YCbCr' : NA_Types[2:] }

def SafeCastCheck(module_in, NSPtypein, module_out, NSPtypeout):
    """Is the range of the input type a subset of the range of the output type?

       This is false if a C cast can yield garbage.

       The following casts may lose precision:
           Int32 or UInt32 --> Float32 or Complex32
           Int64 or UInt64 --> Float64 or Complex64
    """
    if module_in == MODULE.NUMARRAY:
        NA_TypeCheck(NSPtypein)
    if module_out == MODULE.NUMARRAY:
        NA_TypeCheck(NSPtypeout)
    if module_in == MODULE.PIL:
        PIL_ModeCheck(NSPtypein)
    if module_out == MODULE.PIL:
        PIL_ModeCheck(NSPtypeout)

    if module_in == MODULE.NUMARRAY and module_out == MODULE.NUMARRAY:
        good = NA_UpwardsTo_NA[NSPtypein]
        if NSPtypeout not in good:
            raise UnsafeCast, 'cast from %s to %s is unsafe' % \
              (NA_TypeName(NSPtypein), NA_TypeName(NSPtypeout))
    if module_in == MODULE.NUMARRAY and module_out == MODULE.PIL:
        good = NA_UpwardsTo_PIL[NSPtypein]
        if NSPtypeout not in good:
            raise Unsafecast, 'cast from %s to PIL mode "%s" is unsafe' % \
              (NA_TypeName(NSPtypein), NSPtypeout)
    if module_in == MODULE.PIL and module_out == MODULE.NUMARRAY:
        good = PIL_UpwardsTo_NA[NSPtypein]
        if NA_GetType(NSPtypeout) not in good:
            raise UnsafeCast, 'cast from PIL mode "%s" to %s is unsafe' % \
                  (NSPtypein, NA_TypeName(NSPtypeout))
    if module_in == MODULE.PIL and module_out == MODULE.PIL:
        good = PIL_UpwardsTo_PIL[NSPtypein]
        if NSPtypeout not in good:
            raise UnsafeCast, 'cast from PIL mode "%s" to PIL mode "%s" is unsafe' % \
              (NSPtypein, NSPtypeout)

#> === Standardized parameters =======================================
"""
The types of objects that we are dealing with have their own attributes.
Unfortunately, there is no consistent pattern in the names of these
attributes. The most important examples are:

  Standardized      PIL                     numarray

  height, width     size = (width, height)  shape = (rows, columns)
  bands             bands = len(mode)       length of an array in
                                            the third dimension; shape[3]
  NPtype            mode                    Ntype

What saves the situation is that these four attributes are the ones that are
needed most of the time.

These standardized parameters can be accessed one at a time, by function
calls, or all at once creating an instance of class "Parameters". The
functions are Module(arrin), TypeCode(arrin), Width(arrin), Height(arrin),
Bands(arrin), Mode(arrin), NatypeOrMode(arrin), and BytesPerItem(arrin). The
attributes of Parameter objects are module, Ntype, width, height, bands,
mode, NPtype, items and bytes_per_item. For non-PIL arrays, mode
is None. For non-numarrays, Ntype is None.
"""
# Given an array, returns MODULE.PIL or MODULE.NUMARRAY.
def Module(arrin):
    """Returns the module for an array."""
    if isinstance(arrin, Image.Image):
        return MODULE.PIL
    elif isinstance(arrin, numarray.NumArray):
        return MODULE.NUMARRAY
    else:
        raise TypeError, '"arrin" must be a numarray or a PIL image'

# def ModuleName(arrin) See above

def Natype(arrin):
    """Given an numarray, return its Ntype. Otherwise returns None"""
    module = Module(arrin)
    if module == MODULE.PIL:
        return None
    elif module == MODULE.NUMARRAY:
        return arrin.type()

def NatypeOrMode(arrin):
    """Given an array, return its NPtype."""
    module = Module(arrin)
    if module == MODULE.PIL:
        return arrin.mode
    elif module == MODULE.NUMARRAY:
        return arrin.type()

def Mode(arrin):
    """Returns the mode of a PIL image. Otherwise None."""
    if isinstance(arrin, Image.Image):
        return arrin.mode
    return None

def Width(arrin):
    """Returns the width of an array."""
    module = Module(arrin)
    if module == MODULE.PIL:
        return arrin.size[0]
    if module == MODULE.NUMARRAY:
        return arrin.shape[-1]

def Height(arrin):
    """Returns the height of an array."""
    module = Module(arrin)
    if module == MODULE.PIL:
        return arrin.size[1]
    if module == MODULE.NUMARRAY:
        return arrin.shape[-2]

def Bands(arrin):
    """Returns the number of bands of an array."""
    module = Module(arrin)
    if module == MODULE.PIL:
        return PIL_BandsPerMode[arrin.mode]
    if module == MODULE.NUMARRAY:
        if len(arrin.shape) == 2:
            return 1
        elif len(arrin.shape) == 3:
            return arrin.shape[0]
        else:
            raise ValueError, 'numarray must have dimension 2 or 3'

def Elements(arrin):
    """Returns the number of elements in an array."""
    return Width(arrin) * Height(arrin) * Bands(arrin)

def BytesPerItem(arrin):
    """Returns the number of bytes in each element of an array."""
    module = Module(arrin)
    if module == MODULE.PIL:
        return PIL_ModeToType[arrin.mode].bytes
    elif module == MODULE.NUMARRAY:
        return arrin.type().bytes

class Parameters:
    """Calculates parameters for an array.

    Parameters is a class. Its instances have the following attributes:
        module         MODULE.NUMARRAY or MODULE.PIL.
        width          Width of image.
        height         Height of image.
        bands          Number of color planes for image.
        Ntype          If the module is MODULE.NUMARRAY, this is
                       arrin.type().
                       If the module is MODULE.PIL, this is None.
        mode           If the module is MODULE.NUMARRAY, this is None.
                       If the module is MODULE.PIL, this is the mode.
        NPtype         If the module is MODULE.NUMARRAY, this is
                       arrin.type(). If the module is MODULE.PIL, this is
                       the mode.
        items          height * width * bands.
        bytes_per_item The number of bytes per value. For Int32 it is 4.
        totalbytes     items * bytes_per_item.
    """
    def __init__(self, arrin):
        module = Module(arrin)
        if module == MODULE.PIL:
            self.width, self.height = arrin.size
            self.module = MODULE.PIL
            self.Ntype = None
            self.mode = arrin.mode
            self.NPtype = self.mode
            self.bands = PIL_BandsPerMode[arrin.mode]
            self.items = self.width * self.height * self.bands
            self.bytes_per_item = PIL_ModeToType[arrin.mode].bytes
            self.totalbytes = self.items * self.bytes_per_item
        elif module == MODULE.NUMARRAY:
            if len(arrin.shape) == 2:
                # Remember that numarray uses (rows, columns).
                self.height, self.width = arrin.shape
                self.bands = 1
            elif len(arrin.shape) == 3:
                # Remember that numarray uses (rows, columns).
                self.bands, self.height, self.width = arrin.shape
            else:
                raise ValueError, 'if "arrin" is a numarray, ' \
                  'it must have dimensions 2 or 3, "arrin" has dimension %i.' % \
                              len(arrin.shape)
            self.module = MODULE.NUMARRAY
            self.Ntype = arrin.type()
            self.mode = None
            self.NPtype = self.Ntype
            self.items = self.width * self.height * self.bands
            self.bytes_per_item = self.Ntype.bytes
            self.totalbytes = self.items * self.bytes_per_item

    def __repr__(self):
        return '<Parameters %s (%i, %i, %i), %s>' % \
          (ModuleName(self.module), self.bands, self.width, self.height,
            self.NPtype)

#> === Open and Save files =======================================

def New(modulein, typein, shape, value=0):
    if modulein == MODULE.PIL:
        return Image.new(typein, (shape[1], shape[0]), value)
    elif modulein == MODULE.NUMARRAY:
        return numarray.zeros(shape, typein) + value
    else:
        raise TypeError('"modulein" must be MODULE.PIL or MODULE.NUMARRAY')

def Open(filename, NStype=None, bands=None):
    """Opens an image. Returns a numarray."""
    if bands not in (None, 1, 3):
        raise ValueError, 'bands must be None, 1, or 3'
    im = Image.open(filename)
    im.getpixel((0,0))  # Loads image.
    if im.mode == 'P':
        im = im.convert()
    if bands is None:
        if len(im.mode) == 1:
            bands = 1
        else:
            bands = 3
    if bands == 1:
        im = im.convert('L')
    else:
        im = im.convert('RGB')
    if NStype is None:
        return ArrayToArrayCastUpwards(im, MODULE.NUMARRAY)
    else:
        return ArrayToArrayCast(im, MODULE.NUMARRAY, NStype)

def Save(array, filename, mode=None):
    """ Saves an array by clipping and casting to unsigned char (UInt8)."""
    if not IsArray(array):
        raise TypeError, 'tried to Save non-array object.'
    NA_CheckShape(array)
    if len(array.shape) == 2 or \
               (len(array.shape) == 3 and array.shape[0] == 1):
        if len(array.shape) == 3:
            array = array[0]
        if mode is None:
            mode = 'L'
        PILim = NA_To_PIL_Convert(array, mode)
    elif len(array.shape) == 3:
        if mode is None:
            mode = 'RGB'
        PILim = NA_To_PIL_Convert(array, mode)
    else:
        raise ValueError, '"bands" must be 1 or 3.'
    PILim.save(filename)

#> === General conversions ============================================

def NA_To_NA_Convert(array, NStype):
    """Converts one numarray into another. Results are clipped."""
    NA_CheckShape(array)
    Ntype = array.type()
    Ntypeout = NA_GetType(NStype)
    if isinstance(Ntype, ComplexType) or isinstance(Ntypeout, ComplexType):
        raise ComplexNotImplemented, 'Complex types not implemented'
    return clipcast.clipcast(array, Ntypeout)

ClipCast = NA_To_NA_Convert

def PIL_To_PIL_Convert(im, mode):
    """Converts one PIL image into another. Results are clipped."""
    if im.mode == 'F' and mode =='I':
        # PIL returns garbage instead of clipping if float too big.
        array = PIL_To_NA_Convert(im, Int32)
        return NA_To_PIL_Convert(array, mode)
    else:
        return im.convert(mode)

def NA_To_PIL_Convert(array, mode):
    """Converts a numarray into a PIL image. Results are clipped."""
    NA_CheckShape(array)
    Ntype = array.type()
    if isinstance(Ntype, ComplexType):
        raise ComplexNotImplemented, 'Complex types not implemented'
    if mode == '1':
        raise PilModeOneNotImplemented, 'PIL mode "1" is not implemented'
    if mode not in PIL_Modes:
        raise ValueError, 'second argument not a legal PIL mode'
    if not array.iscontiguous():
        array2 = array.copy()
    else:
        array2 = array
    origshape = array2.shape
    if len(array2.shape) == 3 and array2.shape[0] == 1:
        array2.shape = array2.shape[1:]
    if len(array2.shape) == 2:
        height, width = array2.shape
        typeout = PIL_ModeToType[mode]
        array2 = clipcast.clipcast(array, typeout)
        if array2.type() != typeout:
             raise TypeError, 'array2.type(): %s, typeout: %s' % \
                   (array2.type(), typeout)
        # TODO: This stuff should be in C.
        s = array2.tostring()
        tmode = NA_To_PIL_1band[typeout]
        # Note order of sizes.
        im = Image.fromstring(tmode, (width, height), s)
        if mode != im.mode:
            im = im.convert(mode)
        if array2.iscontiguous():
            array2.shape = origshape
            array = array2
        else:
            array = array2.copy()
            array.shape = origshape
        return im
    elif len(array.shape) == 3:
        # The output modes can be "RGB", "RGBA", "CMYK", or "YCbCr".
        Ntype = array.type()
        typeout = UInt8
        bands, height, width = array.shape
        if bands not in (3,4):
            raise ValueError, \
                 'if len(array.shape) == 3, array.shape[0] must be 3 or 4'
        array2 = clipcast.clipcast(array, typeout)
        ims = []
        # TODO: This stuff should be in C.
        for i in range(bands):
            s = array2[i].tostring()
            im = Image.fromstring('L', (width, height), s)
            ims.append(im)
        if bands == 3 or PIL_BandsPerMode[mode] <= 3:
            outim = Image.merge('RGB', ims)
            if PIL_BandsPerMode[mode] < 3:
                outim = outim.convert(mode)
        else:
            # "CYMK" or "YCbCr".
            outim = Image.merge(mode, ims)
        return outim
    else:
        raise ValueError, 'dimension of array must be 2, 3 or 4'

def PIL_To_NA_Convert(im, NStype):
    """Converts a PIL image into a numarray. Results are clipped."""
    if im.mode == '1':
        raise PilModeOneNotImplemented, 'PIL mode "1" is not implemented'
    width, height = im.size
    bands = PIL_BandsPerMode[im.mode]
    typeout = PIL_ModeToType[im.mode]
    s = im.tostring()

    if bands == 1:
        array = numarray.fromstring(s, typeout, (height, width))
    else:
        # For PIL, bands is the most rapidly changing index.
        array = numarray.fromstring(s, typeout, (height, width, bands))
        # Transpose bands.
        perm = range(len(array.shape))
        perm = tuple(perm[-1:] + perm[:-1])
        array = numarray.transpose(array, perm)

    array = clipcast.clipcast(array, NStype)
    return array

def ArrayToArrayCast(arrin, module, NSPtype):
    """Converts one array into another. Results are clipped."""
    pars = Parameters(arrin)
    if pars.module == module == MODULE.PIL and \
                  pars.mode == NSPtype:
        return arrin
    if pars.module == MODULE.NUMARRAY and module == MODULE.NUMARRAY:
        return NA_To_NA_Convert(arrin, NSPtype)
    if pars.module == MODULE.PIL and module == MODULE.PIL:
        return PIL_To_PIL_Convert(arrin, NSPtype)
    if pars.module == MODULE.NUMARRAY and module == MODULE.PIL:
        return NA_To_PIL_Convert(arrin, NSPtype)
    if pars.module == MODULE.PIL and module == MODULE.NUMARRAY:
        return PIL_To_NA_Convert(arrin, NSPtype)

def ArrayToArrayCastUpwards(arrin, module, NSPtypeout=None):
    """Converts one array into another. Information is preserved.

    If "NSPtypeout" is None, the data may be cast upwards into a data
    type that takes more bytes. If "NSPtypeout" is given, the cast will
    be made only into that type or mode if the cast is safe. Casting
       Int32 or UInt32 --> Float32
       Int64 or UInt64 --> Float64
    may lose precision. Nevertheless, we assume that these casts are safe.
    If no safe cast is possible, an exception is raised.
    """
    pars = Parameters(arrin)
    if (pars.module == MODULE.PIL and pars.NPtype == '1') or \
       (module == MODULE.PIL and NSPtypeout == '1'):
        raise PilModeOneNotImplemented, 'Conversions to or from PIL mode "1" are not implemented'
    if NSPtypeout is None:
        if pars.module == MODULE.PIL and module == MODULE.PIL:
            NA_CheckShape(arrin)
            return arrin.copy()
        elif pars.module == MODULE.NUMARRAY and module == MODULE.NUMARRAY:
            return arrin.copy()
        elif pars.module == MODULE.NUMARRAY and module == MODULE.PIL:
            NA_CheckShape(arrin)
            if not NA_To_PIL_1band.has_key(pars.Ntype):
                raise UnsafeCast, 'cannot safely cast numarray of type "%s" ' \
                  'to a PIL image.' % Ntype
            outmode = NA_To_PIL_1band[pars.Ntype]
            if outmode == 'L':
                if bands == 1:
                    mode = 'L'
                elif bands == 3:
                    mode = 'RGB'
                else:
                    raise TypeError, 'if "module" is MODULE.PIL, ' \
                      '"arrin" has type UInt8, and ' \
                      '"NSPtypeout" is None, then "arrin" ' \
                      'must have 1 or 3 bands.'
            return ArrayToArrayCast(arrin, module, outmode)
        elif pars.module == MODULE.PIL and module == MODULE.NUMARRAY:
            if not PIL_ModeToType.has_key(pars.mode):
                raise UnsafeCast, 'cannot safely PIL image of mode "%s" ' \
                  'to a numarray.' % mode
            typeout = PIL_ModeToType[pars.mode]
            return ArrayToArrayCast(arrin, module, typeout)
    else:
        SafeCastCheck(pars.module, pars.NPtype, module,
               NSPtypeout)
        return ArrayToArrayCast(arrin, module, NSPtypeout)

#> === Other functions =============================================

def Split(arrin):
    """Transposes shape (height, width, bands) to (bands, width, height).

    Behaves like split in PIL.
    """
    module = Module(arrin)
    if module != MODULE.NUMARRAY:
        raise TypeError, 'numarray required'
    if Bands(arrin) == 1:
        return arrin
    # Want perm to be (2,0,1).
    perm = range(len(shape))
    perm = tuple(perm[-1:] + perm[:-1])
    return numarray.transpose(arrin, perm)

def Merge(arrin):
    """Transposes shape (bands, height, width) to (width, height, bands).

    Behaves like merge in PIL.
    """
    module = Module(arrin)
    if module != MODULE.NUMARRAY:
        raise TypeError, 'numarray required'
    if len(arrin.shape) <= 2:
        return arrin
    # Want perm to be (1,2,0).
    perm = range(len(shape))
    perm = tuple(perm[1:] + perm[:1])
    return numarray.transpose(arrin, perm)

# Yield all the files in or below "directory". Exclude UNIX hidden files,
# hidden directories and their contents, and all files with extensions in
# the list "excludes". Include only files with extensions in the list
# "includes". Case is ignored for the includes and excludes. If an extension
# is in both lists, it is excluded.
def Walk(top, includes=None, excludes=set()):
    includes2 = None
    if includes != None:
        includes2 = set()
        for include in includes:
            includes2.add(include.lower())
        includes = includes2
    excludes2 = set
    for exclude in excludes:
        excludes2.add(exclude.lower())
    excludes = excludes2
    for thisdir, dirs, files in os.walk(top):
        bad = []
        for i, name in enumerate(dirs):
            if name[0] == '.' or name.find('donotindex') != -1:
                bad.append(i)
        bad.reverse()
        for i in bad:
            del dirs[i]

        files.sort()
        for name in files:
            fullname = os.path.abspath(os.path.join(thisdir, name))
            head, tail = os.path.split(fullname)
            junk, ext = os.path.splitext(fullname)
            ext = ext.lower()
            extOK = False
            if includes == None or ext in includes:
                extOK = True
            if name[-1] != '~' and tail[0] != '.' \
                    and not ext in excludes and extOK:
                yield fullname

# Yield all image files in or below "directory". Exclude UNIX hidden files,
# hidden directories and their contents, and all files with extensions in
# the list "excludes". Include only files with extensions in the list
# "includes". "includes" defaults to the standard image extensions.Case is
# ignored for the includes and excludes. If an extension is in both lists,
# it is excluded.
def WalkImages(top, includes=None, excludes=[], include_standard=True):
    # These are the extensions that can be read by PIL.
    standard_includes = ('.bmp', '.dib', '.cur', '.dcx', '.eps', '.ps',
         '.fli', '.flc', '.fpx', '.gbr', '.gd', '.gif', '.ico', '.im',
         '.jpg', '.jpe', '.jpeg', '.mic', '.msp', '.pcd', '.pcx', '.pdf',
         '.png', '.pbm', '.pgm', '.ppm', '.psd', '.bw', '.rgb', '.cmyk',
         '.ras', '.tga', '.tif', '.tiff', '.xbm', '.xpm')
    if include_standard:
        if includes is None:
            includes = standard_includes
        else:
            includes = tuple(includes) + standard_includes
    for filename in Walk(top, includes, excludes):
        yield filename

def Store(obj, filename, compressed=True):
    s = cPickle.dumps(obj, cPickle.HIGHEST_PROTOCOL)
    if compressed:
        s = zlib.compress(s, 9)
    file(filename, 'w').write(s)

def Retrieve(filename, compressed=True):
    s = file(filename, 'r').read()
    if compressed:
        s = zlib.decompress(s)
    return cPickle.loads(s)

#> === print small arrays ============================================

class PRINTFORMAT:
    UNSIGNED = 0
    SIGNED = 1
    FLOAT = 2

class PRINTSTYLE:
    DEC = 0
    HEX = 1

def Print2D(arrin, format=PRINTFORMAT.UNSIGNED, style=PRINTSTYLE.DEC):
    """Printable form of 2-dim array. Array must be small.

    "format" must be PRINTFORMAT.SIGNED, PRINTFORMAT.UNSIGNED or
    PRINTFORMAT.FLOAT.
    "style" must be PRINTSTYLE.DEC or PRINTSTYLE.HEX.
    """
    if len(arrin.shape) == 1:
        arrin.shape = (1,) + arrin.shape
    if len(arrin.shape) != 2:
        raise TypeError, 'input must have only one band.'

    if arrin.shape[0] > 20 or arrin.shape[1] > 10:
        raise ValueError, 'array too large'

    pars = Parameters(arrin)
    if style == PRINTSTYLE.HEX:
        if format in (PRINTFORMAT.UNSIGNED, PRINTFORMAT.SIGNED):
            f = '%0*x'
            spacing = 4
        else:
            raise ValueError, 'cannot combine PRINTFORMAT.FLOAT and PRINTSTYLE.HEX'
#            f = '%*a'  # Did this ever work? I doubt it.
#            spacing = 9
    else:
        if format in (PRINTFORMAT.UNSIGNED, PRINTFORMAT.SIGNED):
            f = '%*i'
            spacing = 3
        else:
            f = '%*.2f'
            spacing = 6

    out = []
#    out.append('   ')
#    for x in range(pars.width):
#        out.append('%*d' % (spacing, x))
#    out.append('\n')
    for r in range(pars.height):
#        out.append('%3d' % r)
        for c in range(pars.width):
            try:
                out.append(f % (spacing, arrin[r][c]))
            except:
                print >> sys.stderr, 'Error at row %i, col %i' % (r, c)
                raise
        out.append('\n')
    out.append('\n')
    text = ''.join(out)
    return text

def Print3D(arrin, format=PRINTFORMAT.UNSIGNED, style=PRINTSTYLE.DEC):
    """Printable form of 3-dim array. Array must be small.

    "format" must be PRINTFORMAT.SIGNED, PRINTFORMAT.UNSIGNED or PRINTFORMAT.FLOAT.
    "style" must be PRINTSTYLE.DEC or PRINTSTYLE.HEX.
    """
    if len(arrin.shape) != 3:
        raise TypeError, 'input must have exactly three bands.'

    if arrin.shape[1] > 20 or arrin.shape[2] > 10:
        raise ValueError, 'array too large'

    if style == PRINTSTYLE.HEX:
        if format in (PRINTFORMAT.UNSIGNED, PRINTFORMAT.SIGNED):
            f = '%0*x'
            spacing = 4
        else:
            f = '%*a'
            spacing = 9
    else:
        if format in (PRINTFORMAT.UNSIGNED, PRINTFORMAT.SIGNED):
            f = '%*i'
            spacing = 4
        else:
            f = '%*.2f'
            spacing = 6

    pars = Parameters(arrin)
    out = []
    for d in range(pars.bands):
        for r in range(pars.height):
            for c in range(pars.width):
                try:
                    out.append(f % (spacing, arrin[d,r,c]))
                except:
                    print 'Error at band %i, row %i, col %i' % (d, r, c)
                    raise
            out.append('\n')
        out.append('\n')
    out.append('\n')
    text = ''.join(out)
    return text

def PrintArray(arrin):
    """Printable form of 2 or 3 dimensional array. Array must be small."""
    if arrin.shape[-2] > 20 or arrin.shape[-1] > 10:
        raise ValueError, 'array too large'
    pars = Parameters(arrin)
    if pars.Ntype is None:
        raise TypeError, 'input must be numarray'
    if pars.Ntype in (Bool, Complex32, Complex64):
        raise TypeError, 'Bool or Complex images forbidden'
    if pars.Ntype in (Int8, Int16, Int32, Int64):
        format = PRINTFORMAT.SIGNED
    if pars.Ntype in (UInt8, UInt16, UInt32, UInt64):
        format = PRINTFORMAT.UNSIGNED
    if pars.Ntype in (Float32, Float64):
        format = PRINTFORMAT.FLOAT
    style = PRINTSTYLE.DEC

    if Bands(arrin) == 1:
        return Print2D(arrin, format, style)
    else:
        return Print3D(arrin, format, style)

def printbytes(arr, period=None, style=PRINTSTYLE.HEX):
    """Text form for the bytes in an array.

    Because I use a little-endian machine, everything should be read from
    right to left.
    """
    if style == PRINTSTYLE.HEX:
        f = ' %02x'
    else:
        f = ' %3d'
    if type(arr) is not StringType:
        arr = Array_ToString(arr)
    if len(arr) == 0:
        return ''
    out = []
    out.append(f % ord(arr[-1]))
    for i in range(len(arr)-2, -1, -1):
        if period != None and (i+1)%period == 0:
            out.append('\n')
        out.append(f % ord(arr[i]))
    out.append('\n')
    text = ''.join(out)
    return text
