#! /usr/bin/env python

"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, math, numarray
from numarray.numerictypes import *

"""2-d Haar transform.

This is a brute force approach for debugging the algorithm.

"arr" is a Float64 numarray of shape (haarsize, haarsize).
"mask" is an Int32 numarray of the same shape. The pixels to be masked
out have value "1", the others, "0".

A pixel is "bad" if it is masked out, otherwise "good". When summing a group
of pixels, treat each bad pixel as if it equalled the average of the good
pixels. Let m and g be the number of pixels and the number of good pixels,
respectively. Let S be the sum over the good pixels. Result = m*S/g.
"""
def counts(h):
    # Coefficient 0 is the sum of h numbers. Coefficient 1 is X - Y
    # where X and Y are the sums of h//2 numbers. Coefficients 2 and 3
    # have the form X - Y where X and Y are the sums of h//4 terms. ...
    # For h = 16, counts returns
    #       [16, 8, 4, 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1]
    cts = h * [0]
    cts[0] = h
    i = 1
    n = 1
    while h > 1:
        h //= 2
        lo = 0
        for j in range(n):
            cts[i] = h
            lo += 2*h
            i += 1
        n *= 2
    return cts

def sqrt_factor(h):
    """Output array of square roots used by the Haar transform"""
    cts = counts(h)
    out = [math.sqrt(cts[0])]
    for i in range(1, h):
        out.append(math.sqrt(2 * cts[i]))
    return numarray.outerproduct(out, out).astype(Float64)

def exponent(h):
    d = dict()
    for i in range(11):
        d[2**i] = i + 1
    return d[h]

def whiches(h):
    # Returns a dict whose keys are (r, c) and items are
    # (rwin, cwin, rlo, rhi, clo, chi).
    w = dict()
    for i in range(11):
        for j in range(11):
            if i == j == 0:
                w[(i,j)] = (h, h,  0, 1, 0, 1)
            elif i == 0:
                w[(i,j)] = (h, h//2**(j-1),   0, 1, 2**(j-1), 2**j)
            elif j == 0:
                w[(i,j)] = (h//2**(i-1), h,  2**(i-1), 2**i, 0, 1)
            else:
                w[(i,j)] = (h//2**(i-1), h//2**(j-1),
                               2**(i-1), 2**i, 2**(j-1), 2**j)
    return w

def BlockSum(arr, blockshape):
    """Sum over blocks in an array.

    Simplified version of BlockReduce in Numplus.
    """
    dims = len(arr.shape)
    if len(blockshape) != dims:
        raise TypeError, 'blockshape must be same length as shape'
    for i in range(dims):
        arr = arr.copy()
        # Add a new dimension of size step.
        if blockshape[i] > 1:
            newshape = (arr.shape[0] // blockshape[i], blockshape[i]) + arr.shape[1:]
            arr.shape = newshape
            arr = numarray.sum(arr, 1)
        dims = len(arr.shape)
        # (0,1,2,3) --> (1,2,3,0)
        perm =  tuple(range(1, dims)) + (0,)
        arr = numarray.transpose(arr, perm)
    return arr

def Sum(array):
    """Sum over entire array"""
    for n in range(len(array.shape)):
        array = numarray.sum(array)
    return array

def normalize(alist, malist, pixels):
    ma2list = []
    for ma in malist:
        ma2 = numarray.where(ma == pixels, 1, 0)
        ma2list.append(ma2)
    x = numarray.array(ma2list, Int32)
    maskout = numarray.sometrue(x)

    out = []
    for i in range(len(alist)):
        # Prevent divide by zero.
        denom = pixels - malist[i]
        denom = numarray.where(denom == 0, 1.0, denom)
        a = alist[i] * pixels / denom
        a = numarray.where(maskout, 0.0, a)
        out.append(a)
    return out, maskout

def haarstage0(arr, mask):
    h = arr.shape[0]
    pixels = h*h
    ma = BlockSum(mask, (h, h))
    a = BlockSum(arr, (h, h))

    (a,), maskout = normalize((a,), (ma,), pixels)
    return a, maskout

def haarstage1(arr, mask, csize):
    h = arr.shape[0]
    pixels = csize//2 * h
    masksum = BlockSum(mask, (h, csize//2))
    p, q = masksum.shape
    ma = masksum[:, 0:q:2]
    mb = masksum[:, 1:q:2]

    sumarr = BlockSum(arr, (h, csize//2))
    a = sumarr[:, 0:q:2]
    b = sumarr[:, 1:q:2]

    (a2, b2), maskout = normalize((a, b), (ma, mb), pixels)
    return a2 - b2, maskout

def haarstage2(arr, mask, rsize):
    h = arr.shape[0]
    pixels = h * rsize//2
    maskarr = BlockSum(mask, (rsize//2, h))
    p, q = maskarr.shape
    ma = maskarr[0:p:2, :]
    mb = maskarr[1:p:2, :]

    sumarr = BlockSum(arr, (rsize//2, h))
    a = sumarr[0:p:2, :]
    b = sumarr[1:p:2, :]

    (a, b), maskout = normalize((a, b), (ma, mb), pixels)
    return a - b, maskout

def haarstage3(arr, mask, rsize, csize):
    h = arr.shape[0]
    pixels = rsize//2 * csize//2
    maskarr = BlockSum(mask, (rsize//2, csize//2))
    p, q = maskarr.shape
    ma = maskarr[0:p:2, 0:q:2]
    mb = maskarr[0:p:2, 1:q:2]
    mc = maskarr[1:p:2, 0:q:2]
    md = maskarr[1:p:2, 1:q:2]

    sumarr = BlockSum(arr, (rsize//2, csize//2))
    a = sumarr[0:p:2, 0:q:2]
    b = sumarr[0:p:2, 1:q:2]
    c = sumarr[1:p:2, 0:q:2]
    d = sumarr[1:p:2, 1:q:2]

    (a, b, c, d), maskout = normalize((a, b, c, d), (ma, mb, mc, md), pixels)
    return a - b - c + d, maskout

def haar(arr, mask):
    if len(arr.shape) != 2 or arr.shape[0] != arr.shape[1]:
        raise ValueError, 'array must be square and dim = 2'
    if arr.shape != mask.shape:
        raise ValueError, '"arr" and "mask" must have the same shape'  
    h = arr.shape[0]
    arr = numarray.where(mask, 0.0, arr)
    outarr = numarray.zeros(arr.shape, Float64)
    outmask = numarray.zeros(arr.shape, Int32)
    sqrtarr = sqrt_factor(h)
    b = exponent(h)
    w = whiches(h)
    for i in range(b):
        for j in range(b):
            rsize, csize, rlo, rhi, clo, chi = w[(i,j)]
            if i == j == 0:
                result, mask2 = haarstage0(arr, mask)
            elif i == 0:
                result, mask2 = haarstage1(arr, mask, csize)
            elif j == 0:
                result, mask2 = haarstage2(arr, mask, rsize)
            else:
                result, mask2 = haarstage3(arr, mask, rsize, csize)
            outarr[rlo:rhi, clo:chi] = result
            outmask[rlo:rhi, clo:chi] = mask2
    outarr = outarr / sqrtarr
    return outarr, outmask

def test1():
    seq = ((1.0, 3.0), (0.0, 2.0))
    arr = numarray.array(seq, Float64)
    print BlockSum(arr, (2,4))

def test2():
    seq = ((1.0, 3.0), (0.0, 2.0))
#    seq = ((1, 3, 2, -1), (0, -2, -3, 3), (5, 4, 0, -2), (-5, -4, 2, 4))
    arr = numarray.array(seq, Float64)
    mask = numarray.zeros(arr.shape, Int32)
#    mask[0,0] = 1
    outarr, maskarr = haar(arr, mask)
    print maskarr
    print outarr

if __name__ == '__main__':
    numarray.Error.pushMode(all="raise")
    test2()
