
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, datetime, textwrap, shutil, copy
import Image
import utilities, pathnames, mrquery, scorearr, constants
import numarray
from numarray.numerictypes import *
from LN import LN, MLLN

# A global counter that insures that each generated html page has a unique
# url.
unique = 0

# Subscripts for pairdict.
class PD(object):
    N1, N2, ID1, ID2, SCORE = range(5)

#>=== Thumbs =================================================

class thumbs(object):
    """Creates thumbnail images"""
    def __init__(self, mrq):
       self.mrq = mrq
       self.thumbdir = self.mrq.filenames.thumbdir

    def thumb(self, N):
        """Create a thumbnail for html use"""
        virtname = N.virtual_name
        if virtname[0] == '/':
            virtname = virtname[1:]
        thumbname = os.path.join(self.thumbdir, virtname)
        thumbdir = os.path.dirname(thumbname)
        if not os.path.isdir(thumbdir):
            os.makedirs(thumbdir)

        if os.path.isfile(thumbname):
            return thumbname
        try:
            im = Image.open(N)
        except IOError:
            LN(N)
            raise
        if im.mode not in ('L', 'P', 'RGB'):
            raise TypeError('image mode must be "L", "P" or "RGB"')
        if im.mode == 'P':
            im = im.convert('RGB')
        im.thumbnail((constants.THUMB_MAXSIZE, constants.THUMB_MAXSIZE),
                   Image.ANTIALIAS)
        im.save(thumbname, quality=50)
        return thumbname

#>=== Nonce ==================================================

def nonce(pagename, text, prefix):
    """Adds a line of html to my nonce page.

    "nonce word" means "a word coined and used only for a particular
    occasion". The English would say "one off".

    I keep a file on my computer nameed "nonce.html". Whenever a program
    of mine generates some html for me to look at, it puts an url in the
    nonce file.

    The code below is specialized for my particular nonce file format.
    """
    nonce_page = constants.NONCEFILE
    template = """<br><a href="%s">%s</a>\n"""
    html_line = template % (pagename, text)
    f = open(nonce_page, 'r')
    lines = f.read().splitlines()
    f.close()
    linesout = list()
    for line in lines:
        start = line.find('">')
        if start == -1:
            if len(line) > 0:
                linesout.append(line)
            continue
        start += 2
        if not line[start:].startswith(prefix):
            linesout.append(line)
            continue
    linesout.append(html_line)
    linesout.append('')
    out = '\n'.join(linesout)
    open(nonce_page, 'w').write(out)

#>=== HTML ===================================================

def getXtext(topdatadir, basename, dirpair_number):
    """Read an Xtext save file.

    See "html.txt" for an explanation of these files.
    """
    Xtextfile = pathnames.Xtextfile(topdatadir, basename)
    Xtextdict = utilities.Retrieve(Xtextfile)
    return Xtextdict[dirpair_number]

class HTML(object):
    """Creates a tree of html pages"""
    def __init__(self, topdatadir, basename, use_html_piece='normal'):
        self.mrq = mrquery.FetchOrCreate_MRQuery(topdatadir, basename, 'as_is')
        self.topdatadir = topdatadir
        self.basename = basename
        self.use_dirpairs = not (use_html_piece == 'flat')
        self.top_piece = '<html><head><title>mrquery</title></head><body>'
        self.top_piece2 = \
         '<html><head><title>mrquery</title><base target="other"></head><body>'
        self.bottom_piece = '</body</html>\n'

        self.XtextDelete_piece = """\
            X Keep all images.
              Delete first image in each pair of identical images.
              Delete second image in each pair of identical images.
              Delete first image of each non-identical pair.
              Delete second image of each non-identical pair.\n
        """
        self.XtextDelete_piece = textwrap.dedent(self.XtextDelete_piece)

        self.use_html_piece = use_html_piece
        self.pairdict = self.dirpairs()
        self.Xtexts = list()
        self.Xdict = dict()
        self.thumb_maker = thumbs(self.mrq)

        if self.use_html_piece not in ('normal', 'flat'):
            raise ValueError("\"use_html_piece\" must be 'normal' or 'flat'")

        self.kill_old()
        adir = os.path.join(self.mrq.filenames.htmldir, self.use_html_piece)
        if not os.path.isdir(adir):
            os.mkdir(adir)

        if self.use_dirpairs:
            self.top_page()
        else:
            self.top_page_onelist()

    def kill_old(self):
        """Delete any previously generated html pages"""
        names = os.listdir(self.mrq.filenames.htmldir)
        # The names should be 'normal' and/or 'flat'.
        for name in names:
            if name == self.use_html_piece:
                fullname = os.path.join(self.mrq.filenames.htmldir, name)
                shutil.rmtree(fullname)

    def dirpairs(self):
        """Subdivides ddl_sortlist by pairs of directories.

        For any image name, there is the directory "os.dirname(name)".
        Therefore for any (score, imnum1, imnum2) in
        "self.mrq.ddl_sortlist", there is a pair of directories
        (dir1, dir2). These pairs are the keys of "pairdict".
        """
        pairdict = dict()
        # These triples are already sorted by score. Therefore each
        # pairdict[keypair] will be sorted by score.
        for score, N1, N2, ID1, ID2 in self.mrq.ddl_sortlist:
            dir1 = os.path.dirname(N1)
            dir2 = os.path.dirname(N2)
            if dir1 <= dir2:
                keypair = (dir1, dir2)
            else:
                keypair = (dir2, dir1)
            if N1 <= N2:
                valuetuple = (N1, N2, ID1, ID2, score)
            else:
                valuetuple = (N2, N1, ID2, ID1, score)
            # The next line does the same thing as:
            # if keypair not in pairdict:
            #     pairdict[keypair] = list()
            # pairdict[keypair].append(valuetuple)
            pairdict.setdefault(keypair, list()).append(valuetuple)
        return pairdict

    def make_pagename(self, pagename):
        pagename = os.path.join(self.mrq.filenames.htmldir,
                self.use_html_piece, pagename)
        return pagename

    def html_piece(self, N1, N2, ID1, ID2, score, counter):
        """Creates a piece of html comparing two images"""
        template = """\
            <p><font color=green>%i</font> <font color=red>%5.3f</font>
            <br><img src="%s">&nbsp;<img src="%s">
            <br><a href="%s">%s</a> (%i, %i) %s bytes
            <br><a href="%s">%s</a> (%i, %i) %s bytes"""
        template = textwrap.dedent(template)

        thumbname1 = self.thumb_maker.thumb(N1)
        thumbname2 = self.thumb_maker.thumb(N2)
        virtname1 = N1.virtual_name
        virtname2 = N2.virtual_name
        if virtname1.startswith('/'):
            virtname1 = virtname1[1:]
        if virtname2.startswith('/'):
            virtname2 = virtname2[1:]
        name1 = N1.as_str()
        name2 = N2.as_str()
        try:
            return template % (counter, score, thumbname1, thumbname2,
                  name1, virtname1, ID1.rows, ID1.cols, ID1.bytes,
                  name2, virtname2, ID2.rows, ID2.cols, ID2.bytes)
        except TypeError:
            MLLN(ID1, ID2)
            raise

    def html_piece_Xtext(self, N1, N2, ID1, ID2, score, counter):
        """Create a piece of text comparing two images""" 
        # Compare images by size and bytes.
        def cond(IDa, IDb):
            return (IDa.rows > IDb.rows and IDa.cols >= IDb.cols) or \
                   (IDa.rows == IDb.rows and IDa.cols > IDb.cols) or \
                   (IDa.rows == IDb.rows and IDa.cols == IDb.cols)

        name1 = N1.as_str()
        name2 = N2.as_str()
        if os.path.dirname(name1) > os.path.dirname(name2):
            name1, name2 = name2, name1
        if cond(ID1, ID2):
            pair = ('  N %i %5.3f %s' % (counter, score, name1),
                    'X N %i %5.3f %s' % (counter, score, name2))
        elif cond(ID2, ID1):
            pair = ('X N %i %5.3f %s' % (counter, score, name1),
                    '  N %i %5.3f %s' % (counter, score, name2))
        elif ID1 == ID2:
            pair = ('  I %i %5.3f %s' % (counter, score, name1),
                    '  I %i %5.3f %s' % (counter, score, name2))
        else:
            pair = ('  N %i %5.3f %s' % (counter, score, name1),
                    '  N %i %5.3f %s' % (counter, score, name2))
        self.Xtexts.append('\n'.join(pair) + '\n\n')

    def make_pretty(self, N1, N2):
        """Make a pretty name from two filenames"""
        pretty1 = N1.virtual_dirname()
        hashcode1 = self.mrq.name2hash[N1]
        ID1 = self.mrq.hash2ImData[hashcode1]
        pretty2 = N2.virtual_dirname()
        hashcode2 = self.mrq.name2hash[N2]
        ID2 = self.mrq.hash2ImData[hashcode2]
        if ID1.bwbit or ID2.bwbit:
            pretty1 = '<font color=black>%s</font>' % pretty1 
            pretty2 = '<font color=black>%s</font>' % pretty2
        return '%s & %s' % (pretty1, pretty2), (pretty1 == pretty2)

    def html_page(self, quints, counter):
        """Create a page of html"""
        global unique
        out = [self.top_piece2]

        # All the names in quints[k][0] have the same virtual directory.
        # Ditto for quints[k][1].
        N1, N2, ID1, ID2, score = quints[0]
        # N1 and N2 are the pair with the highest score. Use a black font in
        # the pretty names if either of these are b&w.
        pretty, aresame = self.make_pretty(N1, N2)

        virtual_dir1 = os.path.dirname(N1.virtual_name)
        virtual_dir2 = os.path.dirname(N2.virtual_name)
        virtual_dir1 = virtual_dir1[1:].replace('/', '_')
        virtual_dir2 = virtual_dir2[1:].replace('/', '_')
        pagename = """%s.%s.%05i.html""" % (virtual_dir1, virtual_dir2, unique)
        unique += 1
        pagename = self.make_pagename(pagename)

        for N1, N2, ID1, ID2, score in quints:
            if score < constants.CUTOFFSCORE:
                break
            piece = self.html_piece(N1, N2, ID1, ID2, score, counter)
            self.html_piece_Xtext(N1, N2, ID1, ID2, score, counter)
            counter += 1
            if piece is not None:
                out.append(piece)
        out.append(self.bottom_piece)
        text = '\n\n'.join(out)

        open(pagename, 'w').write(text)
        return pagename, pretty, aresame

    def group_page(self, quints):
        """Form groups of 25 image comparisons"""
        global unique

        template1 = """<br><font color=green>%i</font> """ \
                 """<font color=red>%5.3f</font> <a href="%s">%s</a>"""
        template2 = """<br><font color=red>%5.3f</font> <a href="%s">%05i</a>"""
        out = [self.top_piece]
        for i in range(0, len(quints), constants.PAIRSPERPAGE):
            subquints = quints[i : i+constants.PAIRSPERPAGE]
            maxscore = subquints[0][PD.SCORE]
            if maxscore < constants.CUTOFFSCORE:
                break
            pagename, pretty, aresame = self.html_page(subquints, i)
            if self.use_dirpairs:
                piece = template1 % (i, maxscore, pagename, pretty)
            else:
                piece = template2 % (maxscore, pagename, i)
            out.append(piece)
        out.append(self.bottom_piece)
        text = '\n'.join(out)

        # All the names in quints[k][0] have the same virtual directory.
        # Ditto for quints[k][1].
        N1, N2, ID1, ID2, score = quints[0]
        virtual_dir1 = os.path.dirname(N1.virtual_name)
        virtual_dir2 = os.path.dirname(N2.virtual_name)
        virtual_dir1 = virtual_dir1[1:].replace('/', '_')
        virtual_dir2 = virtual_dir2[1:].replace('/', '_')
        pagename = """%s.%s.%05i_group.html""" % (virtual_dir1, virtual_dir2, unique)
        unique += 1
        pagename = self.make_pagename(pagename)
        open(pagename, 'w').write(text)

        # N1 and N2 are the pair with the highest score.
        pretty, aresame = self.make_pretty(N1, N2)
        return pagename, pretty, aresame

    def ignoreme(self, dirpair):
        dirname = pathnames.commondir(dirpair)
        for name in self.mrq.html_ignores:
            if pathnames.isunder(dirname, name):
                return True
        return False

    def top_page(self):
        """Create the top page of html"""
        # Sort by biggest score.
        trips = list()
        for dirpair in self.pairdict:
            maxscore = self.pairdict[dirpair][0][PD.SCORE]
            # Count how many quints are >= constants.UPPERCUTOFFSCORE.
            for i, quint in enumerate(self.pairdict[dirpair]):
                if quint[PD.SCORE] < constants.UPPERCUTOFFSCORE:
                    break
            else:
                # Get here if no break.
                i = len(self.pairdict[dirpair])
            trips.append( (i, maxscore, dirpair) )
        # Sort on number of high scores.
        trips.sort(reverse=True)

        # Cutoff low scores.
        for index in range(len(trips)):
            goodscores, maxscore, dirpair = trips[index]
            if index >= constants.MAXDIRPAIRS or maxscore < constants.CUTOFFSCORE:
                break
        else:
            # Get here if no break.
            index = len(trips)
        trips = trips[:index]

        template = '<br><font color=green>%02i</font> ' \
                   '<font color=brown>%02i</font> ' \
                   '<font color=red>%5.3f</font> %c <a href="%s">%s</a>'

        out = [self.top_piece]
        print 'There are', len(trips), 'directory pairs.'
        ctr = 0
        print '"getXtext.py" can take values in range(%i)' % len(trips)
        for goodscores, maxscore, dirpair in trips:
            if self.ignoreme(dirpair):
                continue
            quints = self.pairdict[dirpair]
            self.Xtexts = list()
            pagelen = len(quints)
            N1, N2, ID1, ID2, score = quints[0]
            if score < constants.CUTOFFSCORE:
                continue
            if (ID1.bwbit or ID2.bwbit) and score < constants.BWCUTOFFSCORE:
                continue
            self.Xtexts.append('\n| %i | %s | %s |\n' % \
                   (ctr, N1.virtual_dirname(), N2.virtual_dirname()))
            self.Xtexts.append(self.XtextDelete_piece)

            if pagelen <= constants.PAIRSPERPAGE:
                pagename, pretty, aresame = self.html_page(quints, 0)
            else:
                pagename, pretty, aresame = self.group_page(quints)
            if aresame:
                code = '.'
            else:
                code = '*'
            piece = template % (ctr, goodscores, maxscore, code, pagename,
                       pretty)
            out.append(piece)

            Xtext = ''.join(self.Xtexts)
            self.Xdict[ctr] = Xtext

            ctr += 1
            if ctr % constants.DIRECTORY_PAIRS_DONE_PRINTED == 0:
                print ctr, 'directory pairs done.'
        out.append(self.bottom_piece)
        text = '\n'.join(out)
        self.write_top_page(text)
        utilities.Store(self.Xdict, self.mrq.filenames.Xtextfile)

    def top_page_onelist(self):
        """Create the top page of html for the 'flat' case"""
        template = """<br><a href="%s">%s</a>"""
        out = [self.top_piece]
        quints = list()
        self.Xtexts.append(self.XtextDelete_piece)
        for index, (score, N1, N2, ID1, ID2) in enumerate(self.mrq.ddl_sortlist):
            if score < constants.CUTOFFSCORE or index > constants.MAXONELISTPAIRS:
                break
            dir1 = os.path.dirname(N1)
            dir2 = os.path.dirname(N2)
            if dir1 <= dir2:
                keypair = (dir1, dir2)
            else:
                keypair = (dir2, dir1)
            if N1 <= N2:
                valuetuple = (N1, N2, ID1, ID2, score)
            else:
                valuetuple = (N2, N1, ID2, ID1, score)
            quints.append(valuetuple)
        pagelen = len(quints)
        if pagelen <= constants.PAIRSPERPAGE:
            pagename, coded, aresame = self.html_page(quints, 0)
        else:
            pagename, coded, aresame = self.group_page(quints)
        piece = template % (pagename, self.basename)
        out.append(piece)
        out.append(self.bottom_piece)
        text = '\n'.join(out)
        self.write_top_page(text)

    def write_top_page(self, text):
        """Write the top page and enter it in 'nonce.html'"""
        dt = datetime.datetime.today()
        datecode = """%02i.%02i.%4i.%02i.%02i.%02i""" % \
          (dt.month, dt.day, dt.year, dt.hour, dt.minute, dt.second)
        basicname = """%s.%s.html""" % (self.basename, datecode)
        top_page_name = self.make_pagename(basicname)
        open(top_page_name, 'w').write(text)
        basicname = '%s.%s' % (self.use_html_piece, basicname)
        bn = '%s.%s' % (self.use_html_piece, self.basename)
        nonce(top_page_name, basicname, bn)

#>=== identicals ===============================================

def identicals(topdatadir, basename):
    """Prepare html for pairs of identical images"""
    print >> sys.stderr, 'Identicals'
    mrq = mrquery.FetchOrCreate_MRQuery(topdatadir, basename, 'as_is')
    dirdict = dict()
    for hashcode in mrq.hash2ImData:
        names = mrq.hash2ImData[hashcode]
        if len(names) < 2:
            continue
        identicals = list()
        for N in names:
            n = N.as_str()
            identicals.append( (os.path.dirname(n), n) )
        identicals.sort()
        identicals = tuple(identicals)
        for dirname, name in identicals:
            if dirname not in dirdict:
                dirdict[dirname] = set()
            dirdict[dirname].add(identicals)

    dirnames = dirdict.keys()
    dirnames.sort()

    outer = list()
    for count, dirname in enumerate(dirnames):
        inner = ['%02i %s' % (count, dirname)]
        print '%02i %s' % (count, dirname)
        for identicals in dirdict[dirname]:
            inner2 = list()
            for dirname, name in identicals:
                line = '<br><a href="%s">%s</a>' % (name, name)
                inner2.append(line)
                print '   ', name
            text = '\n'.join(inner2)
            inner.append(text)
            print
        text = '\n<br>\n'.join(inner)
        outer.append(text)
    outer.append('<br>')
    text = '\n<br><br>\n'.join(outer)
    text = '\n'.join( ('<html><head><title>identicals</title><base target="other"></head><body>',
                    text, '</body></html>', '') )
    dt = datetime.datetime.today()
    datecode = """%02i.%02i.%4i.%02i.%02i.%02i""" % \
      (dt.month, dt.day, dt.year, dt.hour, dt.minute, dt.second)
    basicname = """%s.%s.html""" % (basename, datecode)
    pagename = os.path.join(mrq.filenames.htmldir,
            'identicals', basicname)
    dirname = os.path.join(mrq.filenames.htmldir, 'identicals')
    if not os.path.isdir(dirname):
        os.mkdir(dirname)
    open(pagename, 'w').write(text)
    basicname = 'identicals.%s' % basicname
    bn = 'identicals.%s' % basename
    nonce(pagename, basicname, bn)
