
"""
Copyright 2006 Edward C. Jones

This file is part of mrquery.

mrquery is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

mrquery is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
mrquery; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA 02110-1301 USA
"""
import os, sys, cPickle, time, shutil
from constants import *
import pathnames, clipcast
import Image, numarray
from numarray.numerictypes import *
from LN import LN

def div(num, den):
    """Emulate "div" from C assuming den > 0.

    num  numerator
    den  denominator
    """
    quo, rem = divmod(abs(num), den)
    if num < 0:
        return -quo, -rem
    if num > 0:
        return quo, rem
    return 0, 0

# I constantly end up with hashes, imnums and names in the wrong places. The
# following two classes make debugging easier.
class Hash(str):
    """A thin wrapper around md5 hashcodes"""
    def __getnewargs__(self):
        # Required because of pickling. See "pickletxt.py".
        return (str.__str__(self),)

    def __str__(self):
        return self.hash2hex()

    def __repr__(self):
        return 'Hash(%s)' % str.__repr__(self)

    def hash2ords(self):
        """Convert a hashcode to a list of ints"""
        out = list()
        for c in self:
            out.append(ord(c))
        return tuple(out)

    def hash2hex(self):
        """Convert a hashcode to a hex representation"""
        out = list()
        for c in self:
            s = hex(ord(c))[2:]
            if len(s) == 1:
                s = '0' + s
            out.append(s)
        return ''.join(out)

    def hash2long(self):
        """Convert a hashcode to a llong int"""
        # hashcode is a 16 byte string.
        import struct
        a, b, c, d = struct.unpack('IIII', str.__str__(self))
        p = 2**32
        return ((a*p + b)*p + c)*p + d

class Imnum(int):
    """Very thin wrapper around an int.

    Helps debugging.
    """
    def __getnewargs__(self):
        return (int(self),)

    def __str__(self):
        return 'Imnum(%s)' % int.__str__(self)

    def __repr__(self):
        return 'Imnum(%s)' % int.__repr__(self)

def Store(obj, filename):
    """Pickle an object and write it to a file"""
    s = cPickle.dumps(obj, cPickle.HIGHEST_PROTOCOL)
    file(filename, 'w').write(s)

def Retrieve(filename):
    """Unpickle an object from a file"""
    s = file(filename, 'r').read()
    obj = cPickle.loads(s)
    return obj

def Copy2NewBasename(topdatadir, basename, new_basename):
    """Copies a data directory to a new basename"""
    old_datadir = pathnames.datadirname(topdatadir, basename)
    new_datadir = pathnames.datadirname(topdatadir, new_basename)
    shutil.copytree(old_datadir, new_datadir)    

def redent(text, first, subsequent):
    """Pythonically indents text"""
    lines = text.splitlines()
    sep = '\n' + subsequent * ' '
    text = first * ' ' + sep.join(lines)
    return text.replace(subsequent * ' ' + '\n', '\n')

#>=== Walk ============================================

# Yield all the files in or below "topdir". Exclude UNIX hidden files,
# hidden directories and their contents, and all files with extensions in
# the list "excludes". Include only files with extensions in the list
# "includes". Case is ignored for the includes and excludes. If an extension
# is in both lists, it is excluded. The files are walked in alphabetical
# order. Directories in "excluded_dirs" are not walked.
def Walk(topdirs, includes=None, excludes=None, excluded_dirs=None):
    topdirs = pathnames.prepare_topdirs(topdirs)
    topdirs = list(topdirs)
    topdirs.sort()
    if includes != None:
        includes2 = set()
        for include in includes:
            includes2.add(include.lower())
        includes = includes2
    if excludes is None:
        excludes = tuple()
    excludes2 = set()
    for exclude in excludes:
        excludes2.add(exclude.lower())
    excludes = excludes2
    excluded_dirs = pathnames.prepare_excludes(excluded_dirs)
    for topdir in topdirs:
        for thisdir, dirs, files in os.walk(topdir):
            dirs.sort()
            bad = list()
            for i, name in enumerate(dirs):
                fullname = os.path.join(thisdir, name)
                if name[0] == '.' or fullname in excluded_dirs:
                    # UNIX hidden directory or excluded dir.
                    bad.append(i)
            bad.reverse()
            for i in bad:
                del dirs[i]
    
            files.sort()
            for name in files:
                fullname = os.path.abspath(os.path.join(thisdir, name))
                basename = os.path.basename(fullname)
                junk, ext = os.path.splitext(fullname)
                ext = ext.lower()
                extOK = False
                if includes == None or ext in includes:
                    extOK = True
                if name[-1] != '~' and basename[0] != '.' \
                        and not ext in excludes and extOK:
                    yield fullname

# Yield all image files in or below "directory". Exclude UNIX hidden files,
# hidden directories and their contents, and all files with extensions in
# the list "excludes". Include only files with extensions in the list
# "includes". "includes" defaults to the standard image extensions. Case is
# ignored for the includes and excludes. If an extension is in both lists,
# it is excluded. Directories in "excluded_dirs" are not walked.
def WalkImages(topdirs, includes=None, excludes=None, excluded_dirs=None):
    # These are the extensions that can be read by PIL.
    standard_includes = ('.bmp', '.dib', '.cur', '.dcx', '.eps', '.ps',
         '.fli', '.flc', '.fpx', '.gbr', '.gd', '.gif', '.ico', '.im',
         '.jpg', '.jpe', '.jpeg', '.mic', '.msp', '.pcd', '.pcx', '.pdf',
         '.png', '.pbm', '.pgm', '.ppm', '.psd', '.bw', '.rgb', '.cmyk',
         '.ras', '.tga', '.tif', '.tiff', '.xbm', '.xpm')
    if includes is None:
        includes = standard_includes
    else:
        includes = set(standard_includes) | set(includes)
    for filename in Walk(topdirs, includes, excludes, excluded_dirs):
        yield filename

# Yield all the directories at or below the "topdirs". Exclude UNIX hidden
# directories and their contents. The directories are walked in alphabetical
# order.
def WalkDirs(topdirs):
    topdirs = pathnames.prepare_topdirs(topdirs)
    topdirs = list(topdirs)
    topdirs.sort()
    for topdir in topdirs:
        for thisdir, dirs, files in os.walk(topdir):
            dirs.sort()
            bad = list()
            for i, name in enumerate(dirs):
                fullname = os.path.join(thisdir, name)
                if name[0] == '.':
                    # UNIX hidden directory.
                    bad.append(i)
            bad.reverse()
            for i in bad:
                del dirs[i]
            yield thisdir

#>=== Etc =======================================================

def YIQtoRGB(arr):
    """Convert a YIQ format image to RHB"""
    yiq2rgb = ((1.0, 0.956, 0.621),
               (1.0, -0.272, -0.647),
               (1.0, -1.105, 1.702))
    yiq2rgb = numarray.array(yiq2rgb, Float64)
    arr2 = numarray.transpose(arr, (1,2,0))
    return numarray.innerproduct(yiq2rgb, arr2)

PIL_BandsPerMode = { '1' : 1, 'L' : 1, 'P' : 1, 'I' : 1,
    'F' : 1, 'RGB': 3, 'RGBA': 4, 'CMYK': 4,'YCbCr': 3 }

PIL_ModeToType = { '1' : Bool, 'L' : UInt8, 'P' : UInt8, 'I' : Int32,
    'F' : Float32, 'RGB': UInt8, 'RGBA': UInt8, 'CMYK': UInt8,
    'YCbCr': UInt8 }

def PIL_To_NA_Convert(im, outtype):
    """Converts a PIL image into a numarray. Results are clipped."""
    if im.mode == '1':
        raise NotImplementedError('PIL mode "1" is not implemented')
    width, height = im.size
    bands = PIL_BandsPerMode[im.mode]
    typeout = PIL_ModeToType[im.mode]
    s = im.tostring()

    if bands == 1:
        array = numarray.fromstring(s, typeout, (height, width))
    else:
        # For PIL, bands is the most rapidly changing index.
        array = numarray.fromstring(s, typeout, (height, width, bands))
        # Transpose bands.
        perm = range(len(array.shape))
        perm = tuple(perm[-1:] + perm[:-1])
        array = numarray.transpose(array, perm)

    array = clipcast.clipcast(array, outtype)
    return array
